"""TODO"""
import time
from collections import defaultdict, namedtuple

import numpy as np

import doocspie
from doocspie.doocspie.doocspie_exception import DoocspieException


class Synchronizer:
    _LABEL_KEY = "label"
    _OFFSET_KEY = "offset"

    _Event = namedtuple("Event", ("number", "exception"))
    _Usable = namedtuple("Usable", ("address", "label", "offset"))
    _Unusable = namedtuple("Unusable", ("address", "label", "reason"))

    def __init__(self, properties=None, timeout_seconds=5, buffer_size=32, allow_zero_events=False):
        self._properties = self._conditionally_make_dict(properties)
        self._timeout_seconds = timeout_seconds
        self._buffer_size = buffer_size
        self._allow_zero_events = allow_zero_events
        self._usable_properties, self._unusable_properties, self._zero_event_properties = (), (), ()
        self._actual_properties = ()
        self._buffers = ()
        self._properties_to_update = None
        self._last_event_returned = None
        self.update()

    @staticmethod
    def _conditionally_make_dict(properties):
        if properties is None:
            return {}
        if isinstance(properties, dict):
            return properties.copy()
        if isinstance(properties, (tuple, list)):
            return dict.fromkeys(properties)
        raise DoocspieException("properties must be of type 'dict', 'tuple, or 'list'")

    def _split_properties(self):
        usable_properties, unusable_properties, zero_event_properties = [], [], []
        labels = []
        for address, events in self._get_address_events(self._properties):
            label = self._get_unique_label(address, labels)
            if len(set(events)) == 1:
                event = events[0]
                if event.exception:
                    unusable_properties.append(self._Unusable(address, label, event.exception))
                elif event.number:
                    unusable_properties.append(self._Unusable(address, label, "static event number"))
                else:
                    unusable_properties.append(self._Unusable(address, label, "zero event number"))
                    if self._allow_zero_events:
                        zero_event_properties.append(self._Usable(address, label, 0))
            else:
                usable_properties.append(self._Usable(address, label, self._get_offset(address)))
        return tuple(usable_properties), tuple(unusable_properties), zero_event_properties

    def _get_unique_label(self, address, labels):
        options = self._properties[address]
        if isinstance(options, dict) and self._LABEL_KEY in options:
            label = options[self._LABEL_KEY]
            if label and label in labels:
                raise DoocspieException(self._LABEL_KEY + " '" + label + "' already exists", address)
            labels.append(label)
        else:
            label = None
        return label

    def _get_offset(self, address):
        options = self._properties[address]
        if isinstance(options, dict) and self._OFFSET_KEY in options:
            return options[self._OFFSET_KEY]
        return 0

    def _get_address_events(self, properties):
        address_events = defaultdict(list)
        event_numbers_changed_sufficiently = False
        start = time.time()
        while properties:
            for address in properties:
                try:
                    event = self._Event(doocspie.get(address).event, None)
                except DoocspieException as exc:
                    event = self._Event(np.nan, exc.message)
                address_events[address].append(event)

                if (address_events[address][-1].number - address_events[address][0].number) > 1:
                    event_numbers_changed_sufficiently = True
            if event_numbers_changed_sufficiently:
                break
            if (time.time() - start) > self._timeout_seconds:
                raise DoocspieException("no usable properties found")
        return address_events.items()

    def _get_actual_properties(self):
        actual_properties = []
        for usable_property in self._usable_properties:
            actual_properties.append(usable_property.address)
            if usable_property.label:
                actual_properties.append(usable_property.label)
        for zero_event_property in self._zero_event_properties:
            actual_properties.append(zero_event_property.address)
            if zero_event_property.label:
                actual_properties.append(zero_event_property.label)
        return tuple(actual_properties)

    @property
    def usable_properties(self):
        return self._usable_properties

    @property
    def unusable_properties(self):
        return self._unusable_properties

    @property
    def actual_properties(self):
        return self._actual_properties

    def add(self, address, label=None, offset=None):
        """Add DOOCS address with an optional label to the properties for synchronous readout.

        Args:
            address (str): The DOOCS address to add to the properties for synchronous readout.
            label (str, optional): Label for DOOCS address as a means for alternative property access.
            offset (int, optional): Offset to be applied on event numbers for synchronization.

        Returns:
            None

        Raises:
            DoocspieException: Doocspie related exception for address duplication.
        """
        if address not in self._properties:
            options = {}
            if label:
                options[self._LABEL_KEY] = label
            if offset:
                options[self._OFFSET_KEY] = offset
            self._properties[address] = options
            self._properties_to_update = True
        else:
            raise DoocspieException("address '" + address + "' already exists", address)

    def update(self):
        """Update the newly added properties to for synchronous readout.

        Returns:
            None
        """
        self._usable_properties, self._unusable_properties, self._zero_event_properties = self._split_properties()
        self._actual_properties = self._get_actual_properties()
        self._buffers = tuple({} for _ in range(len(self._usable_properties)))
        self._properties_to_update = False
        self._last_event_returned = None

    def get(self):
        if self._properties_to_update:
            self.update()

        start = time.time()
        while self._usable_properties:
            self._fill_buffers()
            event_matched_readouts = self._find_event_matched_readouts()
            if event_matched_readouts:
                if self._allow_zero_events:
                    self._add_zero_event_readouts(event_matched_readouts)
                return event_matched_readouts
            if (time.time() - start) > self._timeout_seconds:
                raise DoocspieException("synchronization timeout")
        raise DoocspieException("no usable properties found")

    def _fill_buffers(self):
        for usable_property, buffer in zip(self._usable_properties, self._buffers):
            try:
                readout = doocspie.get(usable_property.address)
                event = readout.event + usable_property.offset
                if event not in buffer:
                    if len(buffer) == self._buffer_size:
                        del buffer[list(buffer)[0]]
                    buffer[event] = readout
            except DoocspieException:
                continue

    def _find_event_matched_readouts(self):
        event_matched_readouts = {}
        for event in self._buffers[0]:
            events_in_buffer = [event in buffer for buffer in self._buffers]
            if all(events_in_buffer) and (self._last_event_returned is None or event > self._last_event_returned):
                self._last_event_returned = event
                for index, buffer in enumerate(self._buffers):
                    event_matched_readouts[self._usable_properties[index].address] = buffer[event]
                    if self._usable_properties[index].label:
                        event_matched_readouts[self._usable_properties[index].label] = buffer[event]
                    del buffer[event]
                return event_matched_readouts
        return event_matched_readouts

    def _add_zero_event_readouts(self, readouts):
        for zero_event_property in self._zero_event_properties:
            readouts[zero_event_property.address] = doocspie.get(zero_event_property.address)
            if zero_event_property.label:
                readouts[zero_event_property.label] = readouts[zero_event_property.address]
