"""Module with daq monitor class for simplified readout using DAQ monitors.

This module provides the daq monitor class for simplified readout using the capabilities of DOOCS' DAQ monitors.
"""
import warnings

import doocspie
from doocspie.doocspie.doocspie_exception import DoocspieException


class DaqMonitor:
    """DaqMonitor class for simplified readout using DAQ monitors.

    This class provides the daq monitor for simplified readout using the capabilities of DOOCS' DAQ monitors.
    """
    warnings.formatwarning = lambda msg, *args, **kwargs: f"Warning: {msg}\n"

    def __init__(self, device, facility="FLASH.DAQ"):
        """Constructor of the daq monitor class.

        This constructor initializes the daq monitor using the device and optional facility of the DAQ monitor.

        Args:
            device (str): The DOOCS device address part of the DAQ monitor.
            facility (str, optional): The optional DOOCS facility address part of the DAQ monitor.
        """
        self._device = device
        self._facility = facility
        self._state_address = self._facility + "/" + self._device + "/DAQFSM/DAQ.FSM_STATE_A"
        if not self.is_running():
            warnings.warn(f"DAQ Monitor '{self._facility}/{self._device}' is not running")
        self._address_mapping = self._get_address_mapping()

    def _get_address_mapping(self):
        """Helper method to provide the address mapping between DOOCS property and DAQ monitor address."""
        property_suffix = ".DAQ.SOURCE"
        try:
            addresses = doocspie.ls(self._facility + "/" + self._device + "/MONITOR/*" + property_suffix)
            return {doocspie.get(address).data: address[:-len(property_suffix)] for address in addresses}
        except DoocspieException as exc:
            raise DoocspieException("cannot get address mapping for given DAQ monitor", self._state_address) from exc

    @property
    def properties(self):
        """tuple: The properties available via the DAQ monitor."""
        return tuple(self._address_mapping.keys())

    @classmethod
    def available_monitors(cls, facility="FLASH.DAQ"):
        """List the available DAQ monitors for a given optional facility.

        Returns:
            tuple: The available DAQ monitors for a given optional facility.

        Raises:
            DoocspieException: Doocspie related exception for non-listable available DAQ monitors.
        """
        try:
            return doocspie.ls(facility + "/*.MONITOR")
        except DoocspieException as exc:
            raise DoocspieException("cannot list available DAQ monitors for given facility", facility) from exc

    def is_running(self):
        """Return the DAQ monitor's 'running' state.

        Returns:
            bool:  The 'running' state of the DAQ monitor.

        Raises:
            DoocspieException: Doocspie related exception for inaccessible 'running' state.
        """
        try:
            return doocspie.get(self._state_address).data == "RUN"
        except DoocspieException as exc:
            raise DoocspieException("cannot get 'running' state of given DAQ monitor", self._state_address) from exc

    def refresh(self):
        """Refresh the DAQ monitor in order to see changes of the address mapping.

        Returns:
            None
        """
        if not self.is_running():
            warnings.warn(f"DAQ Monitor '{self._facility}/{self._device}' is not running")
        self._address_mapping = self._get_address_mapping()

    def map(self, address):
        """Map the DOOCS property address to the address of the associated DAQ monitor.

        Args:
            address (str): The DOOCS property address to get the mapped DAQ monitor address for.

        Returns:
            str: The address of the associated DAQ monitor mapped to the given DOOCS property address.

        Raises:
            DoocspieException: Doocspie related exception for non-existing addresses to be mapped.
        """
        if address not in self._address_mapping:
            raise DoocspieException("cannot find mapping for given address", address)
        return self._address_mapping[address]

    def get(self, address, event=None, meta_event=None, start=None, elements=None):
        """Get the readout of a given DOOCS address optional 'event', meta_event', 'start', and 'elements by
        using the associated DAQ monitor.

        Args:
            address (str): The address of the server to get the readout from using the associated DAQ monitor.
            event (int, optional): The optional event to get the data from.
            meta_event (str, optional): The optional Meta property to replace the default event with.
            start (int, optional): The optional start position for readout of 'SPECTRUM' and 'GSPECTRUM' types.
            elements (int, optional): The optional number of elements for readout of 'SPECTRUM' and 'GSPECTRUM' types.

        Returns:
            Readout: The readout object with properties 'data', 'type', 'timestamp', 'event', and 'meta'.

        Raises:
            DoocspieException: It may be related to DOOCS, pydoocs, or doocspie itself.
        """
        try:
            return doocspie.get(self.map(address), event=event, meta_event=meta_event, start=start, elements=elements)
        except DoocspieException as exc:
            raise DoocspieException("cannot get readout for given address", address) from exc
