"""Module with property server class for simplified access to the property servers.

This module provides the property server class for simplified access to property servers provided at FLASH.
"""
import doocspie


class PropertySever:
    """PropertySever class for simplified access to the property servers.

        This class provides the property server class for simplified access to property servers provided at FLASH.
    """
    _FACILITY_DEVICE_PREFIX = "FLASH.EXP/USER.STORE."

    def get_locations(self, full_address=False):
        """Get all available property server locations with the option to return the full addresses.

        Args:
            full_address (bool, optional): The optional state for returning the full addresses.

        Returns:
            tuple: All available property server locations.
        """
        locations = tuple(device.split(".")[-1] for device in doocspie.ls(self._FACILITY_DEVICE_PREFIX + "*"))
        if full_address:
            return tuple(self._FACILITY_DEVICE_PREFIX + location + "/" + location for location in locations)
        return locations

    def get_properties_of_location(self, location, full_address=False):
        """Get all available property server properties with the option to return the full addresses.

        Args:
            location (str): The property server location to request the properties from.
            full_address (bool, optional): The optional state for returning the full addresses.

        Returns:
            tuple: All available property server properties.

        Raises:
            ValueError: Exception for non-existing location.
        """
        if location.startswith(self._FACILITY_DEVICE_PREFIX):
            full_location = location
        else:
            full_location = self._FACILITY_DEVICE_PREFIX + location + "/" + location + "/"

        try:
            properties = tuple(property_.split("/")[-1] for property_ in doocspie.ls(full_location))
        except doocspie.DoocspieException:
            raise ValueError("location '" + location + "' does not exist") from None

        filtered_properties = filter(self._get_relevant, properties)
        if full_address:
            return tuple(full_location + property_ for property_ in filtered_properties)
        return tuple(filtered_properties)

    @staticmethod
    def _get_relevant(property_):
        """Helper method to get only the relevant properties, i.e., those starting with 'VAL' or 'ARRAY'."""
        return property_.startswith("VAL") or property_.startswith("ARRAY")

    def get(self, location, property_):
        """Get the readout from the property server for the given location and property.

        Args:
            location (str): The property server location to request the readout of the property from.
            property_ (str): The actual property to request the readout from.

        Returns:
            Readout: The readout from the property server for the given location and property.

        Raises:
            ValueError: Exception non-existing location and/or property or for not supported full address input.
            DoocspieException: Doocspie related exception for unexpected events.
        """
        self._check_if_full_address(location, property_)
        readout = None
        try:
            readout = doocspie.get(self._FACILITY_DEVICE_PREFIX + location + "/" + location + "/" + property_)
        except doocspie.DoocspieException as exc:
            self._check_valid_location_and_property(location, property_, exc)
        return readout

    def set(self, location, property_, value, allow_resizing=None):
        """Set the content of the property server for the given location and property with the given value.

        Args:
            location (str): The property server location of the property to set the value to.
            property_ (str): The actual property to set the value to.
            value (property dependent): The value to set the property to.
            allow_resizing(bool, optional): The optional state for allow resizing (increasing) array-like DOOCS types.

        Returns:
            None

        Raises:
            ValueError: Exception non-existing location and/or property or for not supported full address input.
            DoocspieException: Doocspie related exception for unexpected events.
        """
        self._check_if_full_address(location, property_)
        try:
            doocspie.set(self._FACILITY_DEVICE_PREFIX + location + "/" + location + "/" + property_, value,
                         allow_resizing)
        except doocspie.DoocspieException as exc:
            self._check_valid_location_and_property(location, property_, exc)

    def _check_if_full_address(self, location, property_):
        """Helper method to check if location and/or property start with the full address."""
        if location.startswith(self._FACILITY_DEVICE_PREFIX) or property_.startswith(self._FACILITY_DEVICE_PREFIX):
            raise ValueError("full address of location and/or property is not supported")

    def _check_valid_location_and_property(self, location, property_, exc):
        """Helper method to check if location and/or property are valid."""
        if location not in self.get_locations():
            raise ValueError("location '" + location + "' does not exist") from None
        if property_ not in self.get_properties_of_location(location):
            raise ValueError("property '" + property_ + "' does not exist") from None
        raise exc
